@@ -664,7 +664,7 @@ int BopomofoRemoveLast(BopomofoData *pBopomofo)
 {
     int i;
 
-    if (pBopomofo->kbtype >= KB_HANYU_PINYIN) {
+    if (BopomofoKbIsPinyin(pBopomofo)) {
         i = strlen(pBopomofo->pinYinData.keySeq);
         pBopomofo->pinYinData.keySeq[i - 1] = '\0';
     } else {
@@ -690,7 +690,7 @@ int BopomofoIsEntering(BopomofoData *pBopomofo)
 {
     int i;
 
-    if (pBopomofo->kbtype >= KB_HANYU_PINYIN) {
+    if (BopomofoKbIsPinyin(pBopomofo)) {
         if (pBopomofo->pinYinData.keySeq[0])
             return 1;
     } else {
@@ -720,4 +720,16 @@ int BopomofoSyllableIndexAlt(BopomofoData *pBopomofo)
 int BopomofoKbType(BopomofoData *pBopomofo)
 {
     return pBopomofo->kbtype;
-}
\ No newline at end of file
+}
+
+int BopomofoKbIsPinyin(BopomofoData *pBopomofo)
+{
+	switch (BopomofoKbType(pBopomofo)) {
+		case KB_HANYU_PINYIN:
+		case KB_THL_PINYIN:
+		case KB_MPS2_PINYIN:
+			return 1;
+		default:
+			return 0;
+	}
+}
