@@ -859,7 +859,7 @@ int MakeOutput(ChewingOutput *pgo, ChewingData *pgdata)
     pgo->chiSymbolCursor = pgdata->chiSymbolCursor;
 
     /* fill bopomofoBuf */
-    if (BopomofoKbType(&pgdata->bopomofoData) >= KB_HANYU_PINYIN) {
+    if (BopomofoKbIsPinyin(&pgdata->bopomofoData)) {
         char key_seq[10];
         BopomofoKeyseq(&pgdata->bopomofoData, key_seq);
         strcpy(pgo->bopomofoBuf, key_seq);
