@@ -1,57 +1,74 @@
 use std::{
-    collections::HashMap,
-    fmt::{Debug, Display},
-    ops::Neg,
-    rc::Rc,
+    collections::{BTreeMap, BTreeSet, VecDeque},
+    fmt::{Debug, Display, Write},
+    iter,
+    ops::{Mul, Neg},
 };
 
-use tracing::trace;
+use log::trace;
 
 use crate::dictionary::{Dictionary, Phrase};
 
-use super::{Break, Composition, ConversionEngine, Glue, Interval, Symbol};
+use super::{Composition, Gap, Interval, Symbol};
 
 /// TODO: doc
-#[derive(Debug)]
+#[derive(Debug, Default)]
 pub struct ChewingEngine;
 
-impl<C: Dictionary + ?Sized> ConversionEngine<C> for ChewingEngine {
-    fn convert(&self, dict: &C, composition: &Composition) -> Vec<Interval> {
-        if composition.buffer.is_empty() {
-            return vec![];
-        }
-        let intervals = self.find_intervals(dict, composition);
-        self.find_best_path(composition.buffer.len(), intervals)
-            .into_iter()
-            .map(|interval| interval.into())
-            .fold(vec![], |acc, interval| glue_fn(composition, acc, interval))
+impl ChewingEngine {
+    const MAX_OUT_PATHS: usize = 100;
+    /// TODO: doc
+    pub fn new() -> ChewingEngine {
+        ChewingEngine
     }
+    pub fn convert<'a>(
+        &'a self,
+        dict: &'a dyn Dictionary,
+        comp: &'a Composition,
+    ) -> impl Iterator<Item = Vec<Interval>> + Clone + 'a {
+        let fast_dp = 'ret: {
+            if comp.is_empty() {
+                break 'ret vec![];
+            }
+            let intervals = self.find_intervals(dict, comp);
+            self.find_best_path(comp.symbols.len(), intervals)
+                .into_iter()
+                .map(|interval| interval.into())
+                .fold(vec![], |acc, interval| glue_fn(comp, acc, interval))
+        };
+        let fast_dp_clone = fast_dp.clone();
+        let mut cached_paths = None;
+        let slow_search = iter::once_with(move || {
+            if comp.is_empty() {
+                return vec![];
+            }
+            cached_paths
+                .get_or_insert_with(|| {
+                    let intervals = self.find_intervals(dict, comp);
+                    let paths = self.find_k_paths(Self::MAX_OUT_PATHS, comp.len(), intervals);
+                    trace!("paths: {:#?}", paths);
+                    debug_assert!(!paths.is_empty());
 
-    fn convert_next(&self, dict: &C, composition: &Composition, next: usize) -> Vec<Interval> {
-        if composition.buffer.is_empty() {
-            return vec![];
-        }
-        let mut graph = Graph::default();
-        let paths = self.find_all_paths(
-            dict,
-            &mut graph,
-            composition,
-            0,
-            composition.buffer.len(),
-            None,
-        );
-        let mut trimmed_paths = self.trim_paths(paths);
-        trimmed_paths.sort();
-        trimmed_paths
-            .into_iter()
-            .rev()
-            .cycle()
-            .nth(next)
-            .map(|p| p.intervals)
-            .expect("should have path")
-            .into_iter()
-            .map(|it| it.into())
-            .fold(vec![], |acc, interval| glue_fn(composition, acc, interval))
+                    let mut trimmed_paths = self.trim_paths(paths);
+                    debug_assert!(!trimmed_paths.is_empty());
+
+                    trimmed_paths.sort_by(|a, b| b.cmp(a));
+                    trimmed_paths
+                })
+                .to_vec()
+        })
+        .flatten()
+        .map(|p| {
+            p.intervals
+                .into_iter()
+                .map(|it| it.into())
+                .fold(vec![], |acc, interval| glue_fn(comp, acc, interval))
+        })
+        // XXX the first result of the slow path should match the fast path.
+        // However the trim path algorithm might remove the first result.
+        .skip_while(move |res| res == &fast_dp);
+
+        iter::once(fast_dp_clone).chain(slow_search)
     }
 }
 
@@ -61,13 +78,15 @@ fn glue_fn(com: &Composition, mut acc: Vec<Interval>, interval: Interval) -> Vec
         return acc;
     }
     let last = acc.last().expect("acc should have at least one item");
-    if com.glues.contains(&Glue(last.end)) {
+    if let Some(Gap::Glue) = com.gap(last.end) {
         let last = acc.pop().expect("acc should have at least one item");
+        let mut phrase = last.str.into_string();
+        phrase.push_str(&interval.str);
         acc.push(Interval {
             start: last.start,
             end: interval.end,
             is_phrase: true,
-            phrase: last.phrase + &interval.phrase,
+            str: phrase.into_boxed_str(),
         })
     } else {
         acc.push(interval);
@@ -75,28 +94,33 @@ fn glue_fn(com: &Composition, mut acc: Vec<Interval>, interval: Interval) -> Vec
     acc
 }
 
-impl ChewingEngine {
-    /// TODO: doc
-    pub fn new() -> ChewingEngine {
-        ChewingEngine
-    }
-}
-
 impl ChewingEngine {
     fn find_best_phrase<D: Dictionary + ?Sized>(
         &self,
         dict: &D,
         start: usize,
         symbols: &[Symbol],
-        selections: &[Interval],
-        breaks: &[Break],
+        com: &Composition,
     ) -> Option<PossiblePhrase> {
         let end = start + symbols.len();
 
-        for br in breaks.iter() {
-            if br.0 > start && br.0 < end {
+        for i in (start..end).skip(1) {
+            if let Some(Gap::Break) = com.gap(i) {
                 // There exists a break point that forbids connecting these
                 // syllables.
+                trace!("No best phrase for {:?} due to break point", symbols);
+                return None;
+            }
+        }
+
+        for selection in &com.selections {
+            if selection.intersect_range(start, end) && !selection.is_contained_by(start, end) {
+                // There's a conflicting partial intersecting selection.
+                trace!(
+                    "No best phrase for {:?} due to selection {:?}",
+                    symbols,
+                    selection
+                );
                 return None;
             }
         }
@@ -108,7 +132,7 @@ impl ChewingEngine {
         let syllables = symbols
             .iter()
             .take_while(|symbol| symbol.is_syllable())
-            .map(|symbol| symbol.as_syllable())
+            .map(|symbol| symbol.to_syllable().unwrap())
             .collect::<Vec<_>>();
         if syllables.len() != symbols.len() {
             return None;
@@ -120,14 +144,14 @@ impl ChewingEngine {
             // If there exists a user selected interval which is a
             // sub-interval of this phrase but the substring is
             // different then we can skip this phrase.
-            for selection in selections.iter() {
-                debug_assert!(!selection.phrase.is_empty());
+            for selection in &com.selections {
+                debug_assert!(!selection.str.is_empty());
                 if start <= selection.start && end >= selection.end {
                     let offset = selection.start - start;
                     let len = selection.end - selection.start;
                     let substring: String =
                         phrase.as_str().chars().skip(offset).take(len).collect();
-                    if substring != selection.phrase {
+                    if substring != selection.str.as_ref() {
                         continue 'next_phrase;
                     }
                 }
@@ -135,30 +159,37 @@ impl ChewingEngine {
 
             // If there are phrases that can satisfy all the constraints
             // then pick the one with highest frequency.
-            if best_phrase.is_none() || phrase.freq() > max_freq {
-                max_freq = phrase.freq();
-                // TODO can we allocate less?
-                best_phrase = Some(phrase.into());
+            if !(phrase.freq() > max_freq || best_phrase.is_none()) {
+                continue;
+            }
+            max_freq = phrase.freq();
+            best_phrase = Some(phrase.into());
+        }
+
+        if best_phrase.is_none() {
+            // try to find if there's a forced selection
+            for selection in &com.selections {
+                if start == selection.start && end == selection.end {
+                    best_phrase = Some(Phrase::new(selection.str.clone(), 0).into());
+                    break;
+                }
             }
         }
 
+        trace!("best phrace for {:?} is {:?}", symbols, best_phrase);
         best_phrase
     }
     fn find_intervals<D: Dictionary + ?Sized>(
         &self,
         dict: &D,
-        comp: &Composition,
+        com: &Composition,
     ) -> Vec<PossibleInterval> {
         let mut intervals = vec![];
-        for begin in 0..comp.buffer.len() {
-            for end in begin..=comp.buffer.len() {
-                if let Some(phrase) = self.find_best_phrase(
-                    dict,
-                    begin,
-                    &comp.buffer[begin..end],
-                    &comp.selections,
-                    &comp.breaks,
-                ) {
+        for begin in 0..com.symbols.len() {
+            for end in begin..=com.symbols.len() {
+                if let Some(phrase) =
+                    self.find_best_phrase(dict, begin, &com.symbols[begin..end], com)
+                {
                     intervals.push(PossibleInterval {
                         start: begin,
                         end,
@@ -213,47 +244,99 @@ impl ChewingEngine {
             .intervals
     }
 
-    fn find_all_paths<'g, D: Dictionary + ?Sized>(
-        &'g self,
-        dict: &D,
-        graph: &mut Graph<'g>,
-        composition: &Composition,
-        start: usize,
-        target: usize,
-        prefix: Option<PossiblePath>,
+    /// Find K possible best alternative paths.
+    ///
+    /// This method is modified from [Yen's algorithm][1] to run on a single
+    /// source and single sink DAG.
+    ///
+    /// [1]: https://en.wikipedia.org/wiki/Yen%27s_algorithm
+    fn find_k_paths(
+        &self,
+        k: usize,
+        end: usize,
+        intervals: Vec<PossibleInterval>,
     ) -> Vec<PossiblePath> {
-        if start == target {
-            return vec![prefix.expect("should have prefix")];
+        let mut ksp = Vec::with_capacity(k);
+        let mut candidates = vec![];
+        let mut graph = Graph::new();
+        for edge in intervals.into_iter() {
+            let values = graph.entry(edge.start).or_default();
+            values.push(edge);
         }
-        let mut result = vec![];
-        for end in start..=target {
-            let entry = graph.entry((start, end));
-            if let Some(phrase) = entry.or_insert_with(|| {
-                self.find_best_phrase(
-                    dict,
-                    start,
-                    &composition.buffer[start..end],
-                    &composition.selections,
-                    &composition.breaks,
-                )
-            }) {
-                let mut prefix = prefix.clone().unwrap_or_default();
-                prefix.intervals.push(PossibleInterval {
-                    start,
-                    end,
-                    phrase: phrase.clone(),
-                });
-                result.append(&mut self.find_all_paths(
-                    dict,
-                    graph,
-                    composition,
-                    end,
-                    target,
-                    Some(prefix),
-                ));
+        ksp.push(
+            self.shortest_path(&graph, &BTreeSet::new(), 0, end)
+                .unwrap(),
+        );
+
+        for kth in 1..k {
+            let prev = kth - 1;
+            for i in 0..ksp[prev].len() {
+                let mut removed_edges = BTreeSet::new();
+                let spur_node = &ksp[prev][i].start;
+                let root_path = &ksp[prev][0..i];
+
+                for p in &ksp {
+                    if root_path.len() < p.len() && root_path == &p[0..i] {
+                        removed_edges.insert((p[i].start, p[i].end));
+                    }
+                }
+
+                if let Some(spur_path) = self.shortest_path(&graph, &removed_edges, *spur_node, end)
+                {
+                    let mut total_path = root_path.to_vec();
+                    total_path.extend(spur_path);
+                    if !ksp.contains(&total_path) {
+                        candidates.push(total_path);
+                    }
+                }
+            }
+            if candidates.is_empty() {
+                break;
+            }
+            candidates.sort_unstable_by_key(|k| k.len());
+            ksp.push(candidates.swap_remove(0));
+        }
+        ksp.into_iter()
+            .map(|intervals| PossiblePath { intervals })
+            .collect()
+    }
+
+    fn shortest_path(
+        &self,
+        graph: &Graph,
+        removed_edges: &BTreeSet<(usize, usize)>,
+        source: usize,
+        sink: usize,
+    ) -> Option<Vec<PossibleInterval>> {
+        let mut parent = vec![None; sink + 1];
+        let mut queue = VecDeque::new();
+        queue.push_back(source);
+        'bfs: while !queue.is_empty() {
+            let node = queue.pop_front().unwrap();
+            if let Some(next_edges) = graph.get(&node) {
+                for edge in next_edges {
+                    if removed_edges.contains(&(edge.start, edge.end)) {
+                        continue;
+                    }
+                    if parent[edge.end].is_none() {
+                        parent[edge.end] = Some(edge);
+                        queue.push_back(edge.end);
+                    }
+                    if edge.end == sink {
+                        break 'bfs;
+                    }
+                }
             }
         }
-        result
+        let mut path = vec![];
+        let mut node = sink;
+        while node != source {
+            let interval = parent[node]?;
+            node = interval.start;
+            path.push(interval.clone());
+        }
+        path.reverse();
+        Some(path)
     }
 
     /// Trim some paths that were part of other paths
@@ -292,16 +375,10 @@ impl ChewingEngine {
 #[derive(Debug, Clone, PartialEq, Eq)]
 enum PossiblePhrase {
     Symbol(Symbol),
-    Phrase(Rc<Phrase>),
+    Phrase(Phrase),
 }
 
 impl PossiblePhrase {
-    fn to_string(&self) -> String {
-        match self {
-            PossiblePhrase::Symbol(sym) => sym.as_char().to_string(),
-            PossiblePhrase::Phrase(phrase) => phrase.to_string(),
-        }
-    }
     fn freq(&self) -> u32 {
         match self {
             PossiblePhrase::Symbol(_) => 0,
@@ -312,13 +389,16 @@ impl PossiblePhrase {
 
 impl Display for PossiblePhrase {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        f.write_str(&self.to_string())
+        match self {
+            PossiblePhrase::Symbol(sym) => f.write_char(sym.to_char().unwrap()),
+            PossiblePhrase::Phrase(phrase) => f.write_str(phrase.as_str()),
+        }
     }
 }
 
 impl From<Phrase> for PossiblePhrase {
     fn from(value: Phrase) -> Self {
-        PossiblePhrase::Phrase(value.into())
+        PossiblePhrase::Phrase(value)
     }
 }
 
@@ -328,13 +408,31 @@ impl From<Symbol> for PossiblePhrase {
     }
 }
 
-#[derive(Clone, Debug, PartialEq, Eq)]
+impl From<PossiblePhrase> for Box<str> {
+    fn from(value: PossiblePhrase) -> Self {
+        match value {
+            PossiblePhrase::Symbol(sym) => sym.to_char().unwrap().to_string().into_boxed_str(),
+            PossiblePhrase::Phrase(phrase) => phrase.into(),
+        }
+    }
+}
+
+#[derive(Clone, PartialEq, Eq)]
 struct PossibleInterval {
     start: usize,
     end: usize,
     phrase: PossiblePhrase,
 }
 
+impl Debug for PossibleInterval {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_tuple("I")
+            .field(&(self.start..self.end))
+            .field(&self.phrase)
+            .finish()
+    }
+}
+
 impl PossibleInterval {
     fn contains(&self, other: &PossibleInterval) -> bool {
         self.start <= other.start && self.end >= other.end
@@ -353,7 +451,7 @@ impl From<PossibleInterval> for Interval {
                 PossiblePhrase::Symbol(_) => false,
                 PossiblePhrase::Phrase(_) => true,
             },
-            phrase: value.phrase.to_string(),
+            str: value.phrase.into(),
         }
     }
 }
@@ -366,7 +464,17 @@ struct PossiblePath {
 impl Debug for PossiblePath {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.debug_struct("PossiblePath")
-            .field("score()", &self.score())
+            .field("rule_largest_sum()", &self.rule_largest_sum().mul(1000))
+            .field(
+                "rule_largest_avgwordlen()",
+                &self.rule_largest_avgwordlen().mul(1000),
+            )
+            .field(
+                "rule_smallest_lenvariance()",
+                &self.rule_smallest_lenvariance().mul(100),
+            )
+            .field("rule_largest_freqsum()", &self.rule_largest_freqsum())
+            .field("total_score()", &self.score())
             .field("intervals", &self.intervals)
             .finish()
     }
@@ -451,7 +559,7 @@ impl PartialEq for PossiblePath {
 
 impl PartialOrd for PossiblePath {
     fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
-        self.score().partial_cmp(&other.score())
+        Some(self.cmp(other))
     }
 }
 
@@ -476,15 +584,13 @@ impl Display for PossiblePath {
     }
 }
 
-type Graph<'a> = HashMap<(usize, usize), Option<PossiblePhrase>>;
+type Graph = BTreeMap<usize, Vec<PossibleInterval>>;
 
 #[cfg(test)]
 mod tests {
-    use std::collections::HashMap;
-
     use crate::{
-        conversion::{Break, Composition, ConversionEngine, Interval, Symbol},
-        dictionary::{Dictionary, Phrase},
+        conversion::{Composition, Gap, Interval, Symbol},
+        dictionary::{Dictionary, Phrase, TrieBuf},
         syl,
         zhuyin::Bopomofo::*,
     };
@@ -492,50 +598,46 @@ mod tests {
     use super::{ChewingEngine, PossibleInterval, PossiblePath};
 
     fn test_dictionary() -> impl Dictionary {
-        HashMap::from([
-            (vec![syl![G, U, O, TONE2]], vec![("國", 1).into()]),
-            (vec![syl![M, I, EN, TONE2]], vec![("民", 1).into()]),
-            (vec![syl![D, A, TONE4]], vec![("大", 1).into()]),
-            (vec![syl![H, U, EI, TONE4]], vec![("會", 1).into()]),
-            (vec![syl![D, AI, TONE4]], vec![("代", 1).into()]),
-            (
-                vec![syl![B, I, AU, TONE3]],
-                vec![("表", 1).into(), ("錶", 1).into()],
-            ),
+        TrieBuf::from([
+            (vec![syl![G, U, O, TONE2]], vec![("國", 1)]),
+            (vec![syl![M, I, EN, TONE2]], vec![("民", 1)]),
+            (vec![syl![D, A, TONE4]], vec![("大", 1)]),
+            (vec![syl![H, U, EI, TONE4]], vec![("會", 1)]),
+            (vec![syl![D, AI, TONE4]], vec![("代", 1)]),
+            (vec![syl![B, I, AU, TONE3]], vec![("表", 1), ("錶", 1)]),
             (
                 vec![syl![G, U, O, TONE2], syl![M, I, EN, TONE2]],
-                vec![("國民", 200).into()],
+                vec![("國民", 200)],
             ),
             (
                 vec![syl![D, A, TONE4], syl![H, U, EI, TONE4]],
-                vec![("大會", 200).into()],
+                vec![("大會", 200)],
             ),
             (
                 vec![syl![D, AI, TONE4], syl![B, I, AU, TONE3]],
-                vec![("代表", 200).into(), ("戴錶", 100).into()],
-            ),
-            (vec![syl![X, I, EN]], vec![("心", 1).into()]),
-            (
-                vec![syl![K, U, TONE4], syl![I, EN]],
-                vec![("庫音", 300).into()],
+                vec![("代表", 200), ("戴錶", 100)],
             ),
+            (vec![syl![X, I, EN]], vec![("心", 1)]),
+            (vec![syl![K, U, TONE4], syl![I, EN]], vec![("庫音", 300)]),
             (
                 vec![syl![X, I, EN], syl![K, U, TONE4], syl![I, EN]],
-                vec![("新酷音", 200).into()],
+                vec![("新酷音", 200)],
             ),
             (
                 vec![syl![C, E, TONE4], syl![SH, TONE4], syl![I, TONE2]],
-                vec![("測試儀", 42).into()],
+                vec![("測試儀", 42)],
             ),
             (
                 vec![syl![C, E, TONE4], syl![SH, TONE4]],
-                vec![("測試", 9318).into()],
+                vec![("測試", 9318)],
             ),
             (
                 vec![syl![I, TONE2], syl![X, I, A, TONE4]],
-                vec![("一下", 10576).into()],
+                vec![("一下", 10576)],
             ),
-            (vec![syl![X, I, A, TONE4]], vec![("下", 10576).into()]),
+            (vec![syl![X, I, A, TONE4]], vec![("下", 10576)]),
+            (vec![syl![H, A]], vec![("哈", 1)]),
+            (vec![syl![H, A], syl![H, A]], vec![("哈哈", 1)]),
         ])
     }
 
@@ -543,54 +645,50 @@ mod tests {
     fn convert_empty_composition() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![],
-            selections: vec![],
-            breaks: vec![],
-            glues: vec![],
-        };
-        assert_eq!(Vec::<Interval>::new(), engine.convert(&dict, &composition));
+        let composition = Composition::new();
+        assert_eq!(
+            Some(Vec::<Interval>::new()),
+            engine.convert(&dict, &composition).next()
+        );
     }
 
     #[test]
     fn convert_simple_chinese_composition() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![
-                Symbol::Syllable(syl![G, U, O, TONE2]),
-                Symbol::Syllable(syl![M, I, EN, TONE2]),
-                Symbol::Syllable(syl![D, A, TONE4]),
-                Symbol::Syllable(syl![H, U, EI, TONE4]),
-                Symbol::Syllable(syl![D, AI, TONE4]),
-                Symbol::Syllable(syl![B, I, AU, TONE3]),
-            ],
-            selections: vec![],
-            breaks: vec![],
-            glues: vec![],
-        };
+        let mut composition = Composition::new();
+        for sym in [
+            Symbol::from(syl![G, U, O, TONE2]),
+            Symbol::from(syl![M, I, EN, TONE2]),
+            Symbol::from(syl![D, A, TONE4]),
+            Symbol::from(syl![H, U, EI, TONE4]),
+            Symbol::from(syl![D, AI, TONE4]),
+            Symbol::from(syl![B, I, AU, TONE3]),
+        ] {
+            composition.push(sym);
+        }
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 2,
                     is_phrase: true,
-                    phrase: "國民".to_string()
+                    str: "國民".into()
                 },
                 Interval {
                     start: 2,
                     end: 4,
                     is_phrase: true,
-                    phrase: "大會".to_string()
+                    str: "大會".into()
                 },
                 Interval {
                     start: 4,
                     end: 6,
                     is_phrase: true,
-                    phrase: "代表".to_string()
+                    str: "代表".into()
                 },
-            ],
-            engine.convert(&dict, &composition)
+            ]),
+            engine.convert(&dict, &composition).next()
         );
     }
 
@@ -598,53 +696,53 @@ mod tests {
     fn convert_chinese_composition_with_breaks() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![
-                Symbol::Syllable(syl![G, U, O, TONE2]),
-                Symbol::Syllable(syl![M, I, EN, TONE2]),
-                Symbol::Syllable(syl![D, A, TONE4]),
-                Symbol::Syllable(syl![H, U, EI, TONE4]),
-                Symbol::Syllable(syl![D, AI, TONE4]),
-                Symbol::Syllable(syl![B, I, AU, TONE3]),
-            ],
-            selections: vec![],
-            breaks: vec![Break(1), Break(5)],
-            glues: vec![],
-        };
+        let mut composition = Composition::new();
+        for sym in [
+            Symbol::from(syl![G, U, O, TONE2]),
+            Symbol::from(syl![M, I, EN, TONE2]),
+            Symbol::from(syl![D, A, TONE4]),
+            Symbol::from(syl![H, U, EI, TONE4]),
+            Symbol::from(syl![D, AI, TONE4]),
+            Symbol::from(syl![B, I, AU, TONE3]),
+        ] {
+            composition.push(sym);
+        }
+        composition.set_gap(1, Gap::Break);
+        composition.set_gap(5, Gap::Break);
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 1,
                     is_phrase: true,
-                    phrase: "國".to_string()
+                    str: "國".into()
                 },
                 Interval {
                     start: 1,
                     end: 2,
                     is_phrase: true,
-                    phrase: "民".to_string()
+                    str: "民".into()
                 },
                 Interval {
                     start: 2,
                     end: 4,
                     is_phrase: true,
-                    phrase: "大會".to_string()
+                    str: "大會".into()
                 },
                 Interval {
                     start: 4,
                     end: 5,
                     is_phrase: true,
-                    phrase: "代".to_string()
+                    str: "代".into()
                 },
                 Interval {
                     start: 5,
                     end: 6,
                     is_phrase: true,
-                    phrase: "表".to_string()
+                    str: "表".into()
                 },
-            ],
-            engine.convert(&dict, &composition)
+            ]),
+            engine.convert(&dict, &composition).next()
         );
     }
 
@@ -652,46 +750,45 @@ mod tests {
     fn convert_chinese_composition_with_good_selection() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![
-                Symbol::Syllable(syl![G, U, O, TONE2]),
-                Symbol::Syllable(syl![M, I, EN, TONE2]),
-                Symbol::Syllable(syl![D, A, TONE4]),
-                Symbol::Syllable(syl![H, U, EI, TONE4]),
-                Symbol::Syllable(syl![D, AI, TONE4]),
-                Symbol::Syllable(syl![B, I, AU, TONE3]),
-            ],
-            selections: vec![Interval {
-                start: 4,
-                end: 6,
-                is_phrase: true,
-                phrase: "戴錶".to_string(),
-            }],
-            breaks: vec![],
-            glues: vec![],
-        };
+        let mut composition = Composition::new();
+        for sym in [
+            Symbol::from(syl![G, U, O, TONE2]),
+            Symbol::from(syl![M, I, EN, TONE2]),
+            Symbol::from(syl![D, A, TONE4]),
+            Symbol::from(syl![H, U, EI, TONE4]),
+            Symbol::from(syl![D, AI, TONE4]),
+            Symbol::from(syl![B, I, AU, TONE3]),
+        ] {
+            composition.push(sym);
+        }
+        composition.push_selection(Interval {
+            start: 4,
+            end: 6,
+            is_phrase: true,
+            str: "戴錶".into(),
+        });
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 2,
                     is_phrase: true,
-                    phrase: "國民".to_string()
+                    str: "國民".into()
                 },
                 Interval {
                     start: 2,
                     end: 4,
                     is_phrase: true,
-                    phrase: "大會".to_string()
+                    str: "大會".into()
                 },
                 Interval {
                     start: 4,
                     end: 6,
                     is_phrase: true,
-                    phrase: "戴錶".to_string()
+                    str: "戴錶".into()
                 },
-            ],
-            engine.convert(&dict, &composition)
+            ]),
+            engine.convert(&dict, &composition).next()
         );
     }
 
@@ -699,29 +796,28 @@ mod tests {
     fn convert_chinese_composition_with_substring_selection() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![
-                Symbol::Syllable(syl![X, I, EN]),
-                Symbol::Syllable(syl![K, U, TONE4]),
-                Symbol::Syllable(syl![I, EN]),
-            ],
-            selections: vec![Interval {
-                start: 1,
-                end: 3,
-                is_phrase: true,
-                phrase: "酷音".to_string(),
-            }],
-            breaks: vec![],
-            glues: vec![],
-        };
+        let mut composition = Composition::new();
+        for sym in [
+            Symbol::from(syl![X, I, EN]),
+            Symbol::from(syl![K, U, TONE4]),
+            Symbol::from(syl![I, EN]),
+        ] {
+            composition.push(sym);
+        }
+        composition.push_selection(Interval {
+            start: 1,
+            end: 3,
+            is_phrase: true,
+            str: "酷音".into(),
+        });
         assert_eq!(
-            vec![Interval {
+            Some(vec![Interval {
                 start: 0,
                 end: 3,
                 is_phrase: true,
-                phrase: "新酷音".to_string()
-            },],
-            engine.convert(&dict, &composition)
+                str: "新酷音".into()
+            },]),
+            engine.convert(&dict, &composition).next()
         );
     }
 
@@ -729,44 +825,45 @@ mod tests {
     fn multiple_single_word_selection() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![
-                Symbol::Syllable(syl![D, AI, TONE4]),
-                Symbol::Syllable(syl![B, I, AU, TONE3]),
-            ],
-            selections: vec![
-                Interval {
-                    start: 0,
-                    end: 1,
-                    is_phrase: true,
-                    phrase: "代".to_string(),
-                },
-                Interval {
-                    start: 1,
-                    end: 2,
-                    is_phrase: true,
-                    phrase: "錶".to_string(),
-                },
-            ],
-            breaks: vec![],
-            glues: vec![],
-        };
+        let mut composition = Composition::new();
+        for sym in [
+            Symbol::from(syl![D, AI, TONE4]),
+            Symbol::from(syl![B, I, AU, TONE3]),
+        ] {
+            composition.push(sym);
+        }
+        for interval in [
+            Interval {
+                start: 0,
+                end: 1,
+                is_phrase: true,
+                str: "代".into(),
+            },
+            Interval {
+                start: 1,
+                end: 2,
+                is_phrase: true,
+                str: "錶".into(),
+            },
+        ] {
+            composition.push_selection(interval);
+        }
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 1,
                     is_phrase: true,
-                    phrase: "代".to_string()
+                    str: "代".into()
                 },
                 Interval {
                     start: 1,
                     end: 2,
                     is_phrase: true,
-                    phrase: "錶".to_string()
+                    str: "錶".into()
                 }
-            ],
-            engine.convert(&dict, &composition)
+            ]),
+            engine.convert(&dict, &composition).next()
         );
     }
 
@@ -774,67 +871,87 @@ mod tests {
     fn convert_cycle_alternatives() {
         let dict = test_dictionary();
         let engine = ChewingEngine::new();
-        let composition = Composition {
-            buffer: vec![
-                Symbol::Syllable(syl![C, E, TONE4]),
-                Symbol::Syllable(syl![SH, TONE4]),
-                Symbol::Syllable(syl![I, TONE2]),
-                Symbol::Syllable(syl![X, I, A, TONE4]),
-            ],
-            selections: vec![],
-            breaks: vec![],
-            glues: vec![],
-        };
+        let mut composition = Composition::new();
+        for sym in [
+            Symbol::from(syl![C, E, TONE4]),
+            Symbol::from(syl![SH, TONE4]),
+            Symbol::from(syl![I, TONE2]),
+            Symbol::from(syl![X, I, A, TONE4]),
+        ] {
+            composition.push(sym);
+        }
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 2,
                     is_phrase: true,
-                    phrase: "測試".to_string()
+                    str: "測試".into()
                 },
                 Interval {
                     start: 2,
                     end: 4,
                     is_phrase: true,
-                    phrase: "一下".to_string()
+                    str: "一下".into()
                 }
-            ],
-            engine.convert_next(&dict, &composition, 0)
+            ]),
+            engine.convert(&dict, &composition).next()
         );
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 3,
                     is_phrase: true,
-                    phrase: "測試儀".to_string()
+                    str: "測試儀".into()
                 },
                 Interval {
                     start: 3,
                     end: 4,
                     is_phrase: true,
-                    phrase: "下".to_string()
+                    str: "下".into()
                 }
-            ],
-            engine.convert_next(&dict, &composition, 1)
+            ]),
+            engine.convert(&dict, &composition).nth(1)
         );
         assert_eq!(
-            vec![
+            Some(vec![
                 Interval {
                     start: 0,
                     end: 2,
                     is_phrase: true,
-                    phrase: "測試".to_string()
+                    str: "測試".into()
                 },
                 Interval {
                     start: 2,
                     end: 4,
                     is_phrase: true,
-                    phrase: "一下".to_string()
+                    str: "一下".into()
                 }
-            ],
-            engine.convert_next(&dict, &composition, 2)
+            ]),
+            engine.convert(&dict, &composition).cycle().nth(2)
+        );
+    }
+
+    #[test]
+    fn convert_pathological_case() {
+        let dict = test_dictionary();
+        let engine = ChewingEngine::new();
+        let mut composition = Composition::new();
+        for _ in 0..80 {
+            composition.push(Symbol::from(syl![H, A]));
+        }
+        assert_eq!(
+            40,
+            engine.convert(&dict, &composition).nth(0).unwrap().len()
+        );
+        assert_eq!(
+            41,
+            engine.convert(&dict, &composition).nth(1).unwrap().len()
+        );
+        assert_eq!(
+            41,
+            engine.convert(&dict, &composition).nth(2).unwrap().len()
         );
     }
 
