@@ -1,23 +1,30 @@
-use std::{path::Path, str};
+use std::{
+    any::Any,
+    error::Error,
+    fmt::Display,
+    path::{Path, PathBuf},
+    str,
+};
 
 use rusqlite::{params, Connection, Error as RusqliteError, OpenFlags, OptionalExtension};
-use thiserror::Error;
 
 use crate::zhuyin::{Syllable, SyllableSlice};
 
 use super::{
-    BuildDictionaryError, DictEntries, Dictionary, DictionaryBuilder, DictionaryInfo,
-    DictionaryUpdateError, Phrase,
+    BuildDictionaryError, Dictionary, DictionaryBuilder, DictionaryInfo, DictionaryMut, Entries,
+    Phrase, UpdateDictionaryError,
 };
 
+const APPLICATION_ID: u32 = 0x43484557; // 'CHEW' in big-endian
+const USER_VERSION: u32 = 0;
+
 /// TODO: doc
-#[derive(Debug, Error)]
-#[error("sqlite error")]
+#[derive(Debug)]
+#[non_exhaustive]
 pub enum SqliteDictionaryError {
     /// TODO: doc
     SqliteError {
         /// TODO: doc
-        #[from]
         source: RusqliteError,
     },
     /// TODO: doc
@@ -28,10 +35,25 @@ pub enum SqliteDictionaryError {
     ReadOnly,
 }
 
+impl Display for SqliteDictionaryError {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(f, "sqlite error")
+    }
+}
+
+impl Error for SqliteDictionaryError {}
+
+impl From<RusqliteError> for SqliteDictionaryError {
+    fn from(value: RusqliteError) -> Self {
+        SqliteDictionaryError::SqliteError { source: value }
+    }
+}
+
 /// TODO: doc
 #[derive(Debug)]
 pub struct SqliteDictionary {
     conn: Connection,
+    path: Option<PathBuf>,
     info: DictionaryInfo,
     read_only: bool,
 }
@@ -39,7 +61,8 @@ pub struct SqliteDictionary {
 impl SqliteDictionary {
     /// TODO: doc
     pub fn open<P: AsRef<Path>>(path: P) -> Result<SqliteDictionary, SqliteDictionaryError> {
-        let mut conn = Connection::open(path)?;
+        let path = path.as_ref().to_path_buf();
+        let mut conn = Connection::open(&path)?;
         Self::initialize_tables(&conn)?;
         Self::migrate_from_userphrase_v1(&mut conn)?;
         Self::ensure_tables(&conn)?;
@@ -47,6 +70,7 @@ impl SqliteDictionary {
 
         Ok(SqliteDictionary {
             conn,
+            path: Some(path),
             info,
             read_only: false,
         })
@@ -56,12 +80,14 @@ impl SqliteDictionary {
     pub fn open_read_only<P: AsRef<Path>>(
         path: P,
     ) -> Result<SqliteDictionary, SqliteDictionaryError> {
-        let conn = Connection::open_with_flags(path, OpenFlags::SQLITE_OPEN_READ_ONLY)?;
+        let path = path.as_ref().to_path_buf();
+        let conn = Connection::open_with_flags(&path, OpenFlags::SQLITE_OPEN_READ_ONLY)?;
         Self::ensure_tables(&conn)?;
         let info = Self::read_info_v1(&conn)?;
 
         Ok(SqliteDictionary {
             conn,
+            path: Some(path),
             info,
             read_only: true,
         })
@@ -76,12 +102,15 @@ impl SqliteDictionary {
 
         Ok(SqliteDictionary {
             conn,
+            path: None,
             info,
             read_only: false,
         })
     }
 
     fn initialize_tables(conn: &Connection) -> Result<(), SqliteDictionaryError> {
+        conn.pragma_update(None, "application_id", APPLICATION_ID)?;
+        conn.pragma_update(None, "user_version", USER_VERSION)?;
         conn.pragma_update(None, "journal_mode", "WAL")?;
         conn.pragma_update(None, "synchronous", "NORMAL")?;
         conn.pragma_update(None, "wal_autocheckpoint", 0)?;
@@ -248,11 +277,11 @@ impl SqliteDictionary {
             let key: String = row.get(0)?;
             let value: String = row.get(1)?;
             match key.as_str() {
-                "name" => info.name = Some(value),
-                "copyright" => info.copyright = Some(value),
-                "license" => info.license = Some(value),
-                "version" => info.version = Some(value),
-                "software" => info.software = Some(value),
+                "name" => info.name = value,
+                "copyright" => info.copyright = value,
+                "license" => info.license = value,
+                "version" => info.version = value,
+                "software" => info.software = value,
                 _ => (),
             }
         }
@@ -260,9 +289,9 @@ impl SqliteDictionary {
     }
 }
 
-impl From<RusqliteError> for DictionaryUpdateError {
+impl From<RusqliteError> for UpdateDictionaryError {
     fn from(source: RusqliteError) -> Self {
-        DictionaryUpdateError {
+        UpdateDictionaryError {
             source: Some(source.into()),
         }
     }
@@ -270,7 +299,7 @@ impl From<RusqliteError> for DictionaryUpdateError {
 
 impl Dictionary for SqliteDictionary {
     fn lookup_first_n_phrases(&self, syllables: &dyn SyllableSlice, first: usize) -> Vec<Phrase> {
-        let syllables_bytes = syllables.get_bytes();
+        let syllables_bytes = syllables.to_bytes();
         let mut stmt = self
             .conn
             .prepare_cached(
@@ -284,7 +313,7 @@ impl Dictionary for SqliteDictionary {
             )
             .expect("SQL error");
         stmt.query_map([syllables_bytes], |row| {
-            let (phrase, freq, time): (String, _, _) = row.try_into()?;
+            let (phrase, freq, time): (Box<str>, _, _) = row.try_into()?;
             let mut phrase = Phrase::new(phrase, freq);
             if let Some(last_used) = time {
                 phrase = phrase.with_time(last_used);
@@ -298,7 +327,7 @@ impl Dictionary for SqliteDictionary {
     }
 
     // FIXME too many clone
-    fn entries(&self) -> Option<DictEntries> {
+    fn entries(&self) -> Entries<'_> {
         let mut stmt = self
             .conn
             .prepare_cached(
@@ -306,9 +335,9 @@ impl Dictionary for SqliteDictionary {
                 FROM dictionary_v1 LEFT JOIN userphrase_v2 ON userphrase_id = id",
             )
             .expect("SQL error");
-        Some(Box::new(
+        Box::new(
             stmt.query_map([], |row| {
-                let (syllables_bytes, phrase, freq, time): (Vec<u8>, String, _, _) =
+                let (syllables_bytes, phrase, freq, time): (Vec<u8>, Box<str>, _, _) =
                     row.try_into()?;
                 let syllables = syllables_bytes
                     .chunks_exact(2)
@@ -329,18 +358,32 @@ impl Dictionary for SqliteDictionary {
             .map(|r| r.unwrap())
             .collect::<Vec<_>>()
             .into_iter(),
-        ))
+        )
     }
 
     fn about(&self) -> DictionaryInfo {
         self.info.clone()
     }
 
-    fn reopen(&mut self) -> Result<(), DictionaryUpdateError> {
+    fn path(&self) -> Option<&Path> {
+        self.path.as_ref().map(|p| p as &Path)
+    }
+
+    fn as_dict_mut(&mut self) -> Option<&mut dyn DictionaryMut> {
+        if !self.read_only {
+            Some(self)
+        } else {
+            None
+        }
+    }
+}
+
+impl DictionaryMut for SqliteDictionary {
+    fn reopen(&mut self) -> Result<(), UpdateDictionaryError> {
         Ok(())
     }
 
-    fn flush(&mut self) -> Result<(), DictionaryUpdateError> {
+    fn flush(&mut self) -> Result<(), UpdateDictionaryError> {
         self.conn.pragma_update(None, "wal_checkpoint", "PASSIVE")?;
         Ok(())
     }
@@ -349,13 +392,13 @@ impl Dictionary for SqliteDictionary {
         &mut self,
         syllables: &dyn SyllableSlice,
         phrase: Phrase,
-    ) -> Result<(), DictionaryUpdateError> {
+    ) -> Result<(), UpdateDictionaryError> {
         if self.read_only {
-            return Err(DictionaryUpdateError {
+            return Err(UpdateDictionaryError {
                 source: Some(Box::new(SqliteDictionaryError::ReadOnly)),
             });
         }
-        let syllables_bytes = syllables.get_bytes();
+        let syllables_bytes = syllables.to_bytes();
         let mut stmt = self.conn.prepare_cached(
             "INSERT OR REPLACE INTO dictionary_v1 (
                     syllables,
@@ -373,13 +416,13 @@ impl Dictionary for SqliteDictionary {
         phrase: Phrase,
         user_freq: u32,
         time: u64,
-    ) -> Result<(), DictionaryUpdateError> {
+    ) -> Result<(), UpdateDictionaryError> {
         if self.read_only {
-            return Err(DictionaryUpdateError {
+            return Err(UpdateDictionaryError {
                 source: Some(Box::new(SqliteDictionaryError::ReadOnly)),
             });
         }
-        let syllables_bytes = syllables.get_bytes();
+        let syllables_bytes = syllables.to_bytes();
         let tx = self.conn.transaction()?;
         {
             let mut stmt = tx.prepare_cached(
@@ -425,8 +468,8 @@ impl Dictionary for SqliteDictionary {
         &mut self,
         syllables: &dyn SyllableSlice,
         phrase_str: &str,
-    ) -> Result<(), DictionaryUpdateError> {
-        let syllables_bytes = syllables.get_bytes();
+    ) -> Result<(), UpdateDictionaryError> {
+        let syllables_bytes = syllables.to_bytes();
         let mut stmt = self
             .conn
             .prepare_cached("DELETE FROM dictionary_v1 WHERE syllables = ? AND phrase = ?")?;
@@ -464,8 +507,8 @@ impl From<RusqliteError> for BuildDictionaryError {
     }
 }
 
-impl From<std::str::Utf8Error> for BuildDictionaryError {
-    fn from(source: std::str::Utf8Error) -> Self {
+impl From<str::Utf8Error> for BuildDictionaryError {
+    fn from(source: str::Utf8Error) -> Self {
         BuildDictionaryError {
             source: Box::new(source),
         }
@@ -478,21 +521,11 @@ impl DictionaryBuilder for SqliteDictionaryBuilder {
         {
             let mut stmt =
                 tx.prepare("INSERT OR REPLACE INTO info_v1 (key, value) VALUES (?, ?)")?;
-            if let Some(name) = info.name {
-                stmt.execute(["name", &name])?;
-            }
-            if let Some(copyright) = info.copyright {
-                stmt.execute(["copyright", &copyright])?;
-            }
-            if let Some(license) = info.license {
-                stmt.execute(["license", &license])?;
-            }
-            if let Some(version) = info.version {
-                stmt.execute(["version", &version])?;
-            }
-            if let Some(software) = info.software {
-                stmt.execute(["software", &software])?;
-            }
+            stmt.execute(["name", &info.name])?;
+            stmt.execute(["copyright", &info.copyright])?;
+            stmt.execute(["license", &info.license])?;
+            stmt.execute(["version", &info.version])?;
+            stmt.execute(["software", &info.software])?;
         }
         tx.commit()?;
         Ok(())
@@ -509,7 +542,7 @@ impl DictionaryBuilder for SqliteDictionaryBuilder {
         } else {
             0
         };
-        let syllables_bytes = syllables.get_bytes();
+        let syllables_bytes = syllables.to_bytes();
         let mut stmt = self.dict.conn.prepare_cached(
             "INSERT OR REPLACE INTO dictionary_v1 (
                     syllables,
@@ -535,6 +568,10 @@ impl DictionaryBuilder for SqliteDictionaryBuilder {
         self.dict.conn.execute("VACUUM INTO ?", [path])?;
         Ok(())
     }
+
+    fn as_any(&self) -> &dyn Any {
+        self
+    }
 }
 
 #[cfg(test)]
@@ -542,10 +579,12 @@ mod tests {
     use std::error::Error;
 
     use rusqlite::{params, Connection};
-    use tempfile::NamedTempFile;
+    use tempfile::{tempdir, NamedTempFile};
 
     use crate::{
-        dictionary::{Dictionary, Phrase},
+        dictionary::{
+            Dictionary, DictionaryBuilder, DictionaryMut, Phrase, SqliteDictionaryBuilder,
+        },
         syl,
         zhuyin::Bopomofo,
     };
@@ -604,6 +643,19 @@ mod tests {
         );
     }
 
+    #[test]
+    fn open_read_only() {
+        let temp_dir = tempdir().expect("Unable to create tempdir");
+        let temp_path = temp_dir.path().join("readonly.sqlite3");
+        let mut builder = SqliteDictionaryBuilder::new();
+        builder.build(&temp_path).expect("Build failure");
+
+        let mut dict =
+            SqliteDictionary::open_read_only(&temp_path).expect("Unable to open database");
+        assert_eq!(temp_path.to_path_buf(), dict.path().unwrap());
+        assert!(dict.as_dict_mut().is_none());
+    }
+
     #[test]
     fn insert_and_update_user_freq() -> Result<(), Box<dyn Error>> {
         let mut dict = SqliteDictionary::open_in_memory()?;
